@@ -14,9 +14,9 @@ from shutil import move
 from struct import Struct
 from sys import modules
 from tempfile import NamedTemporaryFile
-from threading import Semaphore
+from threading import Semaphore, Lock
 from time import sleep
-from typing import IO, ClassVar, Generator, List, Optional, Set, Tuple, Union
+from typing import IO, ClassVar, Dict, Generator, List, Optional, Set, Tuple, Union
 from zlib import compress, crc32, decompress
 
 try:
@@ -36,10 +36,10 @@ class CaptureFile:
 
     If the capture file does not already exist and it is opened for write, or if
     `force_new_empty_file` is True, then a new file will be created and the
-    initial value for its metadata will be the passed `initial_metadata`.
-    These are the only cases where the passed `initial_metadata` is used, and
-    it is provided as a way of optionally ensuring that a capture file always
-    has metadata even when it is first created.
+    initial value for its metadata will be the passed `initial_metadata`. These
+    are the only cases where the passed `initial_metadata` is used, and it is
+    provided as a way of optionally ensuring that a capture file always has
+    metadata even when it is first created.
 
     The `encoding` argument is used to decode records that are returned. The
     default is `utf8`, which means the binary records stored in the capture file
@@ -48,9 +48,26 @@ class CaptureFile:
     encodings available at
     https://docs.python.org/3/library/codecs.html#standard-encodings are valid.
 
-    Only one process can open a capture file for writing at a time. Multiple
-    processes can open the same capture file for read simultaneously with each
-    other and with one process that opens it for write.
+    To ensure only one process can open a capture file for writing at a time set
+    `use_os_file_locking` to True. Multiple processes can always open the same
+    capture file for read simultaneously with each other and with one process
+    that opens it for write.
+
+    Single process but multi-threaded applications do not need
+    `use_os_file_locking` to be True because the CaptureFile module will manage
+    contention using in-memory locks. File locking in some Linux operating/file
+    systems does not work well across servers and even sometimes on a single
+    server so be sure to verify any specific scenario that depends on file
+    locking.
+
+    By default the CaptureFile is tuned for a commit size of approximately 32KB
+    by having the default value of `compression_block_size` set to 32768. Any
+    amount less than this is re-written every commit until the amount of data
+    equals or exceeds this number at which point the data is compressed, written
+    out and (mostly) never re-written again. If commits will typically contain
+    substantially more than 32KB of data, setting `compression_block_size` to a
+    larger number can improve the amount of compression obtained, resulting in a
+    smaller CaptureFile.
 
     An `InvalidCaptureFile` exception is raised if this constructor is used to
     open a file that is not a valid capture file, is in an unsupported version
@@ -70,11 +87,18 @@ class CaptureFile:
     _filenames_opened_for_write: ClassVar[Set[Path]] = set()
     """For in-process double checking to prevent multiple to-write opens."""
 
+    _filenames_with_master_node_lock_sem: ClassVar[Semaphore] = Semaphore()
+    _filenames_with_master_node_lock: ClassVar[
+        Dict[Path, "ReferenceCountedLock"]
+    ] = dict()
+
     file_name: str
-    to_write: InitVar[bool] = False
+    to_write: bool = False
     initial_metadata: InitVar[Optional[bytes]] = None
     force_new_empty_file: InitVar[bool] = False
     encoding: Optional[str] = "utf_8"
+    use_os_file_locking: bool = False
+    compression_block_size: InitVar[int] = 32768
 
     _file_name: Path = field(init=False)
     """A "Path" instance of file_name set during __post_init__"""
@@ -89,6 +113,8 @@ class CaptureFile:
 
     _current_master_node: "MasterNode" = field(init=False)
 
+    _new_is_in_progress: bool = field(init=False)
+
     _record_count: int = field(init=False)
 
     _block_cache: _lru_cache_wrapper = field(init=False)
@@ -96,18 +122,27 @@ class CaptureFile:
 
     def __post_init__(
         self,
-        to_write: bool,
         initial_metadata: Optional[bytes],
         force_new_empty_file: bool,
+        compression_block_size: int,
     ):
         self._block_cache = lru_cache(maxsize=10)(self._block_cache_method)
         self._full_node_cache = lru_cache(maxsize=10)(self._full_node_cache_method)
 
         self._file_name = Path(self.file_name)
 
-        if force_new_empty_file or (to_write and not self._file_name.is_file()):
-            self._new_file(initial_metadata)
-        self.open(to_write)
+        if force_new_empty_file or (self.to_write and not self._file_name.is_file()):
+            self._new_is_in_progress = True
+            self._new_file(initial_metadata, compression_block_size)
+        self._new_is_in_progress = False
+        self.open(self.to_write)
+
+    def __str__(self):
+        if self._file:
+            status = f'opened for {"writing" if self.to_write else "reading"}'
+        else:
+            status = "currently closed but last seen"
+        return f'"{self._file_name}" {status} with {self._record_count:,} records'
 
     def open(self, to_write: bool = False):
         """Opens this CaptureFile instance for reading or writing, depending on
@@ -161,6 +196,15 @@ class CaptureFile:
                     )
                 CaptureFile._filenames_opened_for_write.add(self._file_name)
 
+        with CaptureFile._filenames_with_master_node_lock_sem:
+            if self._file_name not in CaptureFile._filenames_with_master_node_lock:
+                CaptureFile._filenames_with_master_node_lock[
+                    self._file_name
+                ] = ReferenceCountedLock()
+            CaptureFile._filenames_with_master_node_lock[
+                self._file_name
+            ].add_reference()
+
         self._config = CaptureFileConfiguration.read(self._file)
         self.refresh()
 
@@ -177,11 +221,21 @@ class CaptureFile:
             if self.to_write:
                 with CaptureFile._filenames_opened_for_write_sem:
                     CaptureFile._filenames_opened_for_write.remove(self._file_name)
+            if not self._new_is_in_progress:
+                # Master node locks are not used for new files since the
+                # temporary file cannot be in use by any other process
+                with CaptureFile._filenames_with_master_node_lock_sem:
+                    if CaptureFile._filenames_with_master_node_lock[
+                        self._file_name
+                    ].drop_reference():
+                        del CaptureFile._filenames_with_master_node_lock[
+                            self._file_name
+                        ]
 
     def __del__(self):
         self.close()
 
-    def _new_file(self, initial_metadata: Optional[bytes]):
+    def _new_file(self, initial_metadata: Optional[bytes], compression_block_size: int):
         """Creates a new capture file with name `file_name`.
 
         If the file already exists, it is overwritten by the newly created file.
@@ -198,7 +252,9 @@ class CaptureFile:
                 )
             CaptureFile._filenames_opened_for_write.add(self._file_name)
 
-        self._config = CaptureFileConfiguration()
+        self._config = CaptureFileConfiguration(
+            compression_block_size=compression_block_size
+        )
         self._init_compression_block()
 
         # First build the capture file as a temporary file so that we never have
@@ -376,20 +432,21 @@ class CaptureFile:
         self.close()
 
     def _acquire_lock_for_writing(self, /):
-        if "msvcrt" in modules:
-            # we must be on windows
-            lseek(self._file.fileno(), CaptureFile._lock_start_position, SEEK_SET)
-            msvcrt.locking(  # noqa
-                self._file.fileno(), msvcrt.LK_LOCK, CaptureFile._lock_size  # noqa
-            )
-        else:
-            # we are probably on some Unix variant
-            result = fcntl.lockf(
-                self._file.fileno(),
-                fcntl.LOCK_EX | fcntl.LOCK_NB,
-                CaptureFile._lock_size,
-                CaptureFile._lock_start_position,
-            )
+        if self.use_os_file_locking:
+            if "msvcrt" in modules:
+                # we must be on windows
+                lseek(self._file.fileno(), CaptureFile._lock_start_position, SEEK_SET)
+                msvcrt.locking(  # noqa
+                    self._file.fileno(), msvcrt.LK_LOCK, CaptureFile._lock_size  # noqa
+                )
+            else:
+                # we are probably on some Unix variant
+                result = fcntl.lockf(
+                    self._file.fileno(),
+                    fcntl.LOCK_EX | fcntl.LOCK_NB,
+                    CaptureFile._lock_size,
+                    CaptureFile._lock_start_position,
+                )
 
     @contextmanager
     def _acquire_master_nodes_lock(self, /):
@@ -401,35 +458,41 @@ class CaptureFile:
 
     def _acquire_master_nodes_lock_internal(self, lock: bool, /):
         assert self._file
-        lock_size = self._config.master_node_size * 2
-        if "msvcrt" in modules:
-            lseek(self._file.fileno(), self._config.page_size, SEEK_SET)
-            # we must be on windows
-            # added comments below to suppress my-py errors when we are viewing code on Linux
-            lock_mode = msvcrt.LK_LOCK if lock else msvcrt.LK_UNLCK  # type: ignore[attr-defined]
-            msvcrt.locking(self._file.fileno(), lock_mode, lock_size)  # type: ignore[attr-defined]
+        if not self._new_is_in_progress:
+            # Master node locks are not used for new files since the temporary
+            # file cannot be in use by any other process
+            CaptureFile._filenames_with_master_node_lock[self._file_name].lock(lock)
+            if self.use_os_file_locking:
+                lock_size = self._config.master_node_size * 2
+                if "msvcrt" in modules:
+                    # we must be on windows
+                    lseek(self._file.fileno(), self._config.page_size, SEEK_SET)
+                    lock_mode = msvcrt.LK_RLCK if lock else msvcrt.LK_UNLCK  # type: ignore[attr-defined]
+                    # lock_mode = msvcrt.LK_LOCK if lock else msvcrt.LK_UNLCK  # type: ignore[attr-defined]
+                    msvcrt.locking(self._file.fileno(), lock_mode, lock_size)  # type: ignore[attr-defined]
+                    # added line comments above to suppress my-py errors when we are viewing code on Linux
 
-            # On 2021-09-26 discovered that on Windows 10 buffer reads after the
-            # lock have incorrect bytes after the first 4k bytes if read in
-            # partial pages. E.g. read 4 bytes then read more than 4k more
-            # bytes. The bytes that appeared after the 4k were the original 4
-            # bytes read. This only happened if a read from position 0 happened
-            # before the lock. Reading a page from the file after the lock
-            # seemed to fix the issue. Reading more than 4k to start did not
-            # help
-            self._file.seek(self._config.page_size)
-            self._file.read(self._config.page_size)
-        else:
-            # we are probably on some Unix variant
-            # added comments below to suppress my-py errors when we are viewing code on Windows
-            lock_type = fcntl.LOCK_EX if self.to_write else fcntl.LOCK_SH  # type: ignore[attr-defined]
-            lock_mode = lock_type if lock else fcntl.LOCK_UN  # type: ignore[attr-defined]
-            fcntl.lockf(  # type: ignore[attr-defined]
-                self._file.fileno(),
-                lock_mode,
-                lock_size,
-                self._config.page_size,
-            )
+                    # On 2021-09-26 discovered that on Windows 10 buffer reads after the
+                    # lock have incorrect bytes after the first 4k bytes if read in
+                    # partial pages. E.g. read 4 bytes then read more than 4k more
+                    # bytes. The bytes that appeared after the 4k were the original 4
+                    # bytes read. This only happened if a read from position 0 happened
+                    # before the lock. Reading a page from the file after the lock
+                    # seemed to fix the issue. Reading more than 4k to start did not
+                    # help
+                    self._file.seek(self._config.page_size)
+                    self._file.read(self._config.page_size)
+                else:
+                    # we are probably on some Unix variant
+                    lock_type = fcntl.LOCK_EX if self.to_write else fcntl.LOCK_SH  # type: ignore[attr-defined]
+                    lock_mode = lock_type if lock else fcntl.LOCK_UN  # type: ignore[attr-defined]
+                    # added line comments above to suppress my-py errors when we are viewing code on Windows
+                    fcntl.lockf(  # type: ignore[attr-defined]
+                        self._file.fileno(),
+                        lock_mode,
+                        lock_size,
+                        self._config.page_size,
+                    )
 
     def get_metadata(self, /) -> Optional[bytes]:
         """Returns the binary metadata that was stored in the capture file on
@@ -643,35 +706,37 @@ class CaptureFile:
             raise IndexError
         rightmost_nodes = self._current_master_node.rightmost_path.rightmost_nodes
         # Use "reversed" so we start at the root instead of the leaves
-        nodes = reversed(rightmost_nodes)
-        path = reversed(
-            list(
-                compute_path(
-                    # 1 is subtracted from record_number because Python is
-                    # 0-based while CaptureFile records start at 1
-                    record_number - 1,
-                    len(rightmost_nodes),
-                    self._config.fan_out,
-                )
+        root_to_leaf_rightmost_nodes = reversed(rightmost_nodes)
+        root_to_leaf_path = reversed(
+            leaf_to_root_path(
+                # 1 is subtracted from record_number because Python is
+                # 0-based while CaptureFile records start at 1
+                record_number - 1,
+                len(rightmost_nodes),
+                self._config.fan_out,
             )
         )
 
         # skip nodes as long as path follows rightmost nodes.
-        while (child_index := next(path)) == len((node := next(nodes)).children):
-            pass
+        for child_index, current_rightmost_node in zip(
+            root_to_leaf_path, root_to_leaf_rightmost_nodes
+        ):
+            if child_index != len(current_rightmost_node.children):
+                break
 
         # get first persistent child's data cooridnates. This child will refer
         # to either the record or the root of a perfect sub-tree of which no
         # decendant can be a rightmost node of the top level tree.
-        child = node.children[child_index]
+        current_child_coordinates = current_rightmost_node.children[child_index]
 
         # iterate through the remainder of the path of child indexes until we
         # arrive at the data coordinates of the record
-        for child_index in path:
-            child_full_node = self._full_node_cache(child)
-            child = child_full_node[child_index]
+        for child_index in root_to_leaf_path:
+            current_child_coordinates = self._full_node_cache(
+                current_child_coordinates
+            )[child_index]
 
-        return child.record(self)
+        return current_child_coordinates.record(self)
 
     def record_count(self, /) -> int:
         """Returns the number of records available when the file was opened or
@@ -852,7 +917,7 @@ class CaptureFile:
 
         if not self.to_write:
             raise CaptureFileNotOpenForWrite(
-                f'Cannot commit "{self.file_name}" because it is not open for writting.'
+                f'Cannot commit "{self.file_name}" because it is not open for writing.'
             )
 
         self._file.flush()
@@ -937,9 +1002,20 @@ class CaptureFileConfiguration:
         self.full_node_struct = Struct(">" + "QL" * self.fan_out)
 
     @classmethod
-    def read(cls, file, /) -> "CaptureFileConfiguration":
+    def read(cls, file: IO[bytes], /) -> "CaptureFileConfiguration":
         file.seek(0)
-        buffer = file.read(cls.struct.size)
+        for _ in range(30):
+            # Retry every 0.1 seconds for 3 seconds in case the master node lock
+            # interferes with the reading. This interference was observed in
+            # Windows 11 on 2022-06-04 where simply having a master node lock
+            # prevented the first few bytes of the file from being read even
+            # though they are not part of the lock range.
+            try:
+                buffer = file.read(cls.struct.size)
+            except PermissionError:
+                sleep(0.1)
+            else:
+                break
         (
             header,
             version,
@@ -1371,6 +1447,39 @@ class BytesStream(BytesIO):
         self.write(b"\0" * (end_position - self.tell()))
 
 
+@dataclass
+class ReferenceCountedLock:
+    _reference_count: int = 0
+    _lock: Lock = Lock()
+
+    def add_reference(self) -> None:
+        self._reference_count += 1
+
+    def drop_reference(self) -> bool:
+        """Returns True when the last reference has been dropped to indicate
+        this lock is no longer in use and it is time to clear any references to
+        this lock"""
+        self._reference_count -= 1
+        return self._reference_count == 0
+
+    def lock(self, lock: bool) -> None:
+        if lock:
+            self._lock.acquire()
+        else:
+            self._lock.release()
+
+
+def leaf_to_root_path(position: int, height: int, fan_out: int, /) -> List[int]:
+    """Compute the path of child indexes from the leaf through the nodes to the
+    root."""
+
+    path = [0] * height
+    for i in range(height):
+        position, path[i] = divmod(position, fan_out)
+
+    return path
+
+
 class CaptureFileAlreadyOpen(Exception):
     pass
 
@@ -1385,13 +1494,3 @@ class CaptureFileNotOpenForWrite(Exception):
 
 class InvalidCaptureFile(Exception):
     pass
-
-
-def compute_path(position: int, height: int, fan_out: int, /):
-    """Compute the path of child indexes from the leaf through the nodes to the
-    root."""
-
-    while height > 0:
-        height -= 1
-        position, child_index = divmod(position, fan_out)
-        yield child_index
